package cn.com.qmth.stmms.ms.admin.service;

import cn.com.qmth.stmms.ms.admin.dto.StudentDTO;
import cn.com.qmth.stmms.ms.commons.config.ImageCompressionConfig;
import cn.com.qmth.stmms.ms.commons.config.ImageConfig;
import cn.com.qmth.stmms.ms.commons.config.SystemConfig;
import cn.com.qmth.stmms.ms.commons.constant.SystemConstant;
import cn.com.qmth.stmms.ms.commons.utils.MD5Util;
import cn.com.qmth.stmms.ms.commons.utils.RandomUtil;
import cn.com.qmth.stmms.ms.commons.utils.excel.ExcelError;
import cn.com.qmth.stmms.ms.commons.utils.excel.ExcelReader;
import cn.com.qmth.stmms.ms.commons.utils.excel.ExcelReaderHandle;
import cn.com.qmth.stmms.ms.commons.utils.image.ImageCompression;
import cn.com.qmth.stmms.ms.core.domain.ExamQuestion;
import cn.com.qmth.stmms.ms.core.domain.Paper;
import cn.com.qmth.stmms.ms.core.domain.Student;
import cn.com.qmth.stmms.ms.core.repository.ExamQuestionRepo;
import cn.com.qmth.stmms.ms.core.repository.MarkTaskRepo;
import cn.com.qmth.stmms.ms.core.repository.PaperRepo;
import cn.com.qmth.stmms.ms.core.repository.StudentRepo;
import cn.com.qmth.stmms.ms.core.vo.Subject;
import org.apache.commons.codec.digest.DigestUtils;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import javax.imageio.ImageIO;
import java.awt.image.BufferedImage;
import java.io.*;
import java.lang.reflect.Field;
import java.util.Date;
import java.util.List;
import java.util.Objects;
import java.util.Random;

/**
 * 数据上传服务 Created by zhengmin on 2016/11/18.
 */
@Service
public class DataUploadService {
    private static final org.slf4j.Logger LOGGER = LoggerFactory.getLogger(DataUploadService.class);

    @Autowired
    private StudentRepo studentRepo;

    @Autowired
    private ExamQuestionRepo examQuestionRepo;

    @Autowired
    private PaperRepo paperRepo;

    @Autowired
    private SystemConfig systemConfig;

    @Autowired
    private ImageCompressionConfig compressionConfig;

    @Autowired
    private MarkTaskRepo markTaskRepo;

    @Autowired
    ImageConfig imageConfig;

    /**
     * 上传考生试卷数据
     *
     * @param studentId 考生id
     * @param subject   科目
     * @param in        图片文件输入流
     * @param isManual  是否手工绑定
     * @throws IOException
     */
    @Transactional
    public void uploadPaper(Long studentId, Subject subject, InputStream in, boolean isManual) throws Exception {
        long start = System.currentTimeMillis();
        LOGGER.info("准备生成裁切原图和缩略图:{}", start);
        File imageFileTemp = null;
        try {
            FileInputStream sheetIn = null;
            FileInputStream slicein = null;
            Student student = studentRepo.findOne(studentId);
            if (imageConfig.isImageEnc()) {
                //保存裁切原图+文件名加密
                String imageDir = systemConfig.getImageDir() + File.separator + student.getWorkId() + File.separator + subject
                        + File.separator + student.getAreaCode();
                File out = new File(imageDir);
                if (!out.exists()) {
                    out.mkdirs();
                }

                String imageMd5 = MD5Util.getImageRuleMd5(student.getWorkId(), subject.ordinal(), student.getAreaCode(), student.getExamNumber(), student.getId());
                File imageFile = new File(imageDir + File.separator + imageMd5 + ".jpg");
                imageFileTemp = new File(imageDir + File.separator + imageMd5 + "temp.jpg");
                LOGGER.info("image存放目录:{}", imageFile.getPath());
                OutputStream outputStream = new FileOutputStream(imageFile);
                OutputStream outputStreamTemp = new FileOutputStream(imageFileTemp);
                int index = 0;
                byte[] bytes = new byte[1024];
                byte[] bytesEnc = new byte[1024];
                while ((index = in.read(bytes)) != -1) {
                    //将字节数组的数据全部写入到输出流中
                    for (int i = 0; i < index; i++) {
                        //通过异或运算加密
                        bytesEnc[i] = (byte) (bytes[i] ^ SystemConstant.SECRET_KEY);
                    }
                    outputStream.write(bytesEnc, 0, index);
                    outputStreamTemp.write(bytes, 0, index);
                }
                in.close();
                outputStream.flush();
                outputStream.close();
                outputStreamTemp.flush();
                outputStreamTemp.close();

                //生成缩略图
                String thumbDir = systemConfig.getThumbDir() + File.separator + student.getWorkId() + File.separator + subject
                        + File.separator + student.getAreaCode();
                File thumb = new File(thumbDir);
                if (!thumb.exists()) {
                    thumb.mkdirs();
                }

                BufferedImage bufferedImage = ImageCompression.compress(imageFileTemp, compressionConfig);
                File thumbFile = new File(thumbDir + File.separator + imageMd5 + ".jpg");
                LOGGER.info("thumb存放目录:{}", thumbFile.getPath());
                OutputStream outputStream1 = new FileOutputStream(thumbFile);
                ByteArrayOutputStream os = new ByteArrayOutputStream();
                ImageIO.write(bufferedImage, "jpg", os);
                InputStream in1 = new ByteArrayInputStream(os.toByteArray());

                writeStream(in1, outputStream1);
                long end = System.currentTimeMillis();
                LOGGER.info("生成原图和缩略图耗时：{}", (end - start) / 1000 + "s");

                in1.close();
                outputStream1.flush();
                outputStream1.close();
                sheetIn = new FileInputStream(imageFile.getPath());
                slicein = new FileInputStream(thumbFile.getPath());
            } else {
                String savePath = systemConfig.getImageDir() + File.separator + student.getWorkId() + File.separator + subject
                        + File.separator + student.getAreaCode();
                File out = new File(savePath);
                if (!out.exists()) {
                    out.mkdirs();
                }
                File outFile = new File(savePath + File.separator + student.getExamNumber() + ".jpg");
                if (!outFile.exists()) {
                    outFile.createNewFile();
                }
                OutputStream output = new FileOutputStream(outFile);
                byte[] buffer = new byte[in.available()];
                in.read(buffer);
                output.write(buffer);
                output.close();

                // 生成缩略图
                String thumbDir = systemConfig.getThumbDir() + File.separator + student.getWorkId() + File.separator + subject
                        + File.separator + student.getAreaCode();
                File thumb = new File(thumbDir);
                if (!thumb.exists()) {
                    thumb.mkdirs();
                }
                BufferedImage bufferedImage = ImageCompression.compress(outFile, compressionConfig);
                String thumbFileName = thumbDir + File.separator + student.getExamNumber() + ".jpg";
                ImageIO.write(bufferedImage, "jpg", new File(thumbFileName));

                sheetIn = new FileInputStream(savePath + File.separator + student.getExamNumber() + ".jpg");
                slicein = new FileInputStream(thumbFileName);
            }
            Paper exist = paperRepo.findByWorkIdAndSubjectAndExamNumber(student.getWorkId(), subject,
                    student.getExamNumber());
            String sheetMD5 = DigestUtils.md5Hex(sheetIn);
            String sliceMD5 = DigestUtils.md5Hex(slicein);
            sheetIn.close();
            slicein.close();
            if (exist != null) {
                exist.setManual(isManual);
                exist.setUploadedOn(new Date());
                exist.setSheetMD5(sheetMD5);
                exist.setSliceMD5(sliceMD5);
                paperRepo.save(exist);
            } else {
                ExamQuestion examQuestion = examQuestionRepo.findByWorkIdAndSubjectAndAreaCode(student.getWorkId(), subject,
                        student.getAreaCode());
                Long random = getRandom(student.getWorkId(), student.getExamNumber());
                Paper paper = new Paper(student.getWorkId(), null, subject, examQuestion, student, false, random);
                paper.setSheetMD5(sheetMD5);
                paper.setSliceMD5(sliceMD5);
                paper.setExamRoom(student.getExamRoom());
                paperRepo.save(paper);
                // 更新科目上传状态
                /**
                 * Long count =
                 * paperRepo.countByWorkIdAndExamNumber(student.getWorkId(),student.
                 * getExamNumber()); if(count == Subject.values().length){
                 * student.setUpload(true); studentRepo.save(student); }
                 */
                synchronized (this) {
                    String uploadStatus = student.getUploadStatus();
                    String replace = subject.toString() + ":1";
                    String regx = subject.toString() + ":[0-1]";
                    String newStatus = uploadStatus.replaceFirst(regx, replace);
                    student.setUploadStatus(newStatus);
                    student.setAbsent(false);
                    studentRepo.save(student);
                }
            }
        } catch (Exception e) {
            e.printStackTrace();
        } finally {
            //原图删除
            if (Objects.nonNull(imageFileTemp) && imageFileTemp.exists()) {
                imageFileTemp.delete();
            }
        }
    }

    @Transactional
    public List<ExcelError> uploadStudents(Long workId, final boolean isAbsent, InputStream inputStream) {
        ExcelReader excelReader = new ExcelReader(StudentDTO.class);
        List<ExcelError> excelErrors = excelReader.reader(inputStream, new ExcelReaderHandle() {

            @Override
            public ExcelError handle(Object obj) {
                try {
                    StudentDTO dto = (StudentDTO) obj;

                    String examNumber = dto.getExamNumber();
                    //如果没有设置区域代码，用准考证前2位作为区域代码
                    String areaCode = null;
                    if (dto.getAreaCode() == null || dto.getAreaCode().isEmpty()) {
                        areaCode = examNumber.substring(0, 2);
                        dto.setAreaCode(areaCode);
                    }

                    long isExist = examQuestionRepo.countByWorkIdAndAreaCode(workId, dto.getAreaCode());
                    if (isExist == 0) {
                        for (Subject subject : Subject.values()) {
                            ExamQuestion question = new ExamQuestion(dto.getAreaCode(), subject, workId, dto.getAreaName());
                            examQuestionRepo.save(question);
                        }
                    }
                    Student student = studentRepo.findByWorkIdAndExamNumber(workId, dto.getExamNumber());
                    if (student == null && !isAbsent) {
                        student = new Student(dto.getName(), dto.getExamNumber(), dto.getAreaName(), dto.getAreaCode(),
                                workId, dto.getExamRoom(), dto.getSourceName());
                        if (!checkObjFieldIsNotNull(student)) {
                            throw new RuntimeException("考生信息缺失");
                        }
                        studentRepo.save(student);
                    } else if (student != null) {
                        if (isAbsent) {
                            student.setAbsent(isAbsent);
                            student.setUploadStatus("SX:0,SC:0,SM:0");

                            /**
                             * 查询该考生所有paper,若paper无任务则删除，有任务就保留
                             */
                            List<Paper> papers = paperRepo.findByWorkIdAndExamNumber(workId, student.getExamNumber());
                            if (!papers.isEmpty()) {
                                for (Paper paper : papers) {
                                    Long count = markTaskRepo.countByPaperId(paper.getId());
                                    if (count == 0) {
                                        paperRepo.delete(paper);
                                    }
                                }
                            }
                        } else {
                            student.setName(dto.getName());
                            student.setAreaCode(dto.getAreaCode());
                            student.setAreaName(dto.getAreaName());
                            student.setExamRoom(dto.getExamRoom());
                            student.setSourceName(dto.getSourceName());
                        }
                        if (!checkObjFieldIsNotNull(student)) {
                            throw new RuntimeException("考生信息缺失");
                        }
                        studentRepo.save(student);
                    }
                    return null;
                } catch (RuntimeException e) {
                    ExcelError excelError = new ExcelError();
                    excelError.setExcelErrorType(e.getMessage());
                    return excelError;
                }
            }
        });
        return excelErrors;
    }

    /**
     * 清空考生数据 有上传数据无法清空
     *
     * @param workId
     */
    @Transactional
    public void clearStudents(Long workId) {
        Long isExist = paperRepo.countByWorkId(workId);
        if (isExist > 0) {
            throw new RuntimeException("有考生已上传，无法清空");
        }
        studentRepo.deleteByWorkId(workId);
        examQuestionRepo.deleteByWorkId(workId);
    }

    @Transactional
    public Paper savePaper(Student student, Subject subject, boolean isManual) throws Exception {
        ExamQuestion examQuestion = examQuestionRepo.findByWorkIdAndSubjectAndAreaCode(student.getWorkId(), subject, student.getAreaCode());
        Paper paper = paperRepo.findByWorkIdAndSubjectAndExamNumber(student.getWorkId(), subject, student.getExamNumber());
        if (paper == null) {
            Long random = getRandom(student.getWorkId(), student.getExamNumber());
            paper = new Paper(student.getWorkId(), null, subject, examQuestion, student, isManual, random);
        }
        paper.setUploadedOn(new Date());
        paper.setManual(isManual);

        String sheetPath = null;
        String slicePath = null;
        if (imageConfig.isImageEnc()) {
            String imageMd5 = MD5Util.getImageRuleMd5(student.getWorkId(), subject.ordinal(), student.getAreaCode(), student.getExamNumber(), student.getId());
            sheetPath = systemConfig.getSheetDir() + File.separator + student.getWorkId() + File.separator + subject
                    + File.separator + student.getAreaCode() + File.separator + imageMd5 + ".jpg";
            slicePath = systemConfig.getImageDir() + File.separator + student.getWorkId() + File.separator + subject
                    + File.separator + student.getAreaCode() + File.separator + imageMd5 + ".jpg";
        } else {
            sheetPath = systemConfig.getSheetDir() + File.separator + student.getWorkId() + File.separator + subject
                    + File.separator + student.getAreaCode() + File.separator + student.getExamNumber() + ".jpg";
            slicePath = systemConfig.getImageDir() + File.separator + student.getWorkId() + File.separator + subject
                    + File.separator + student.getAreaCode() + File.separator + student.getExamNumber() + ".jpg";
        }
        FileInputStream sheetIn = new FileInputStream(sheetPath);
        FileInputStream slicein = new FileInputStream(slicePath);
        String sheetMD5 = DigestUtils.md5Hex(sheetIn);
        String sliceMD5 = DigestUtils.md5Hex(slicein);
        paper.setSheetMD5(sheetMD5);
        paper.setSliceMD5(sliceMD5);
        paperRepo.save(paper);
        sheetIn.close();
        slicein.close();
        synchronized (this) {
            String uploadStatus = student.getUploadStatus();
            String replace = subject.toString() + ":1";
            String regx = subject.toString() + ":[0-1]";
            String newStatus = uploadStatus.replaceFirst(regx, replace);
            student.setUploadStatus(newStatus);
            studentRepo.save(student);
        }
        return paper;
    }

    /**
     * 获取随机号
     *
     * @param workId
     * @param examNumber
     * @return
     * @throws Exception
     */
    public Long getRandom(Long workId, String examNumber) throws Exception {
        int count = 0, result = 0;
        Long random = 0L;
        while (true) {
            random = RandomUtil.randomMap.get(workId).get(new Random().nextInt(RandomUtil.randomMap.get(workId).size()));
//            random = RandomUtil.randomList.get(new Random().nextInt(RandomUtil.randomList.size()));
            result = paperRepo.countByWorkIdAndExamNumberAndRandomSeq(workId, examNumber, random);
            if (result == 0 && random != Long.parseLong(examNumber.substring(3, examNumber.length()))) {
                break;
            } else {
                count++;
            }
            if (count > 100) {
//                throw new Exception("重复几率较高，建议重新生成随机号");
                RandomUtil.getRandom(workId, true);
            }
        }
        return random;
    }

    /**
     * 校验学生导入所有属性
     *
     * @param obj
     * @return
     */
    private boolean checkObjFieldIsNotNull(Object obj) {
        for (Field field : obj.getClass().getDeclaredFields()) {
            if (Objects.equals("id", field.getName())) {
                continue;
            }
            field.setAccessible(true);
            try {
                if (field.get(obj) == null) {
                    return false;
                }
            } catch (IllegalAccessException e) {
                return false;
            }
        }
        return true;
    }

    /**
     * 流写入
     *
     * @param inputStream
     * @param outputStream
     * @return
     * @throws IOException
     */
    public OutputStream writeStream(InputStream inputStream, OutputStream outputStream) throws IOException {
        int index = 0;
        byte[] bytes = new byte[1024];
        byte[] bytesEnc = new byte[1024];
        while ((index = inputStream.read(bytes)) != -1) {
            //将字节数组的数据全部写入到输出流中
            for (int i = 0; i < index; i++) {
                //通过异或运算加密
                bytesEnc[i] = (byte) (bytes[i] ^ SystemConstant.SECRET_KEY);
            }
            outputStream.write(bytesEnc, 0, index);
        }
        return outputStream;
    }
}
